/*
 *  Copyright (c) 2023 Fraunhofer Institute for Applied Information Technology
 *
 *  This program and the accompanying materials are made available under the
 *  terms of the Apache License, Version 2.0 which is available at
 *  https://www.apache.org/licenses/LICENSE-2.0
 *
 *  SPDX-License-Identifier: Apache-2.0
 *
 */

package org.fraunhofer.fit.extension.extractor;

import com.github.tbouron.SpdxLicense;
import org.eclipse.edc.spi.monitor.Monitor;
import org.fraunhofer.fit.extension.extractor.model.FairResponseModel;
import org.fraunhofer.fit.extension.extractor.model.FairvalidatorModel;
import org.fraunhofer.fit.extension.extractor.model.ValidationResponseModel;

import java.net.HttpURLConnection;
import java.net.URL;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class FairService {

    private final Monitor monitor;

    public FairService(Monitor monitor) {
        this.monitor = monitor;
    }

    public ValidationResponseModel getValidation(FairvalidatorModel fmodel) {

        ValidationResponseModel validationResponseModel = new ValidationResponseModel();
        List<String> msgList = new ArrayList<String>();

        float validationScore = 0;
        boolean validpid = false;
        boolean validDataLink = false;
        FairResponseModel fairResponseModel = null;

        fairResponseModel = this.pidValidation(fmodel.getPid());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
            validpid = true;
        }

        fairResponseModel = this.licenseValidation(fmodel.getLicence());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.publisherValidation(fmodel.getPublisher());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.titleValidation(fmodel.getTitle());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.dateValidation(fmodel.getPublicationDate(), "Publication ");
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.dateValidation(fmodel.getModifiedDate(), "Modification ");
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.descriptionValidation(fmodel.getDescription());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.keywordValidation(fmodel.getKeywords());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.endpointValidation(fmodel.getDataLink());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
            validDataLink = true;
        }

        if (fmodel.getDataAccessLevel() == "" && (validpid || validDataLink)) {
            fmodel.setDataAccessLevel("public");
        } else {
            fmodel.setDataAccessLevel("non-public");
        }

        fairResponseModel = this.dataAccessLevelValidation(fmodel.getDataAccessLevel());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }



        fairResponseModel = this.mimeTypeValidation(fmodel.getMimeType());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.languageValidation(fmodel.getLanguage());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.contactValidation(fmodel.getContactPoint());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.byteSizeValidation(fmodel.getByteSize());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        fairResponseModel = this.versionValidation(fmodel.getVersion());
        msgList.add(fairResponseModel.getMessage());
        if (fairResponseModel.getIsValid()) {
            validationScore++;
        }

        float validationPercentage = (validationScore / Constants.TOTAL_METADATA) * 100;
        validationResponseModel.setValidationScore(validationPercentage);
        validationResponseModel.setValidationScoreDesc("given metadata is " + validationPercentage + "% FAIR.");
        validationResponseModel.setMessage(msgList);
        return validationResponseModel;
    }

    private String getPidValidatorUrl(String pid) {
        // This method get based URl to check validity of PID.
        // According to https://howtofair.dk/how-to-fair/persistent-identifiers/ most widely used PIDs are DOI and Handle, hence this method only support DOI and Handle PID.

        String baseurl = null;

        Pattern doiPattern = Pattern.compile("(https://doi.org/|doi:\\[)(10\\.\\w+/.+)");
        Pattern hdlPattern = Pattern.compile("(hdl:)?(\\w+\\.?\\w+/.+)");
        Matcher doiMatcher = doiPattern.matcher(pid);
        Matcher hdlMatcher = hdlPattern.matcher(pid);
        if (doiMatcher.find()) {
            baseurl = "https://doi.org/" + doiMatcher.group(2);
        } else if (hdlMatcher.find()) {
            baseurl =  "http://hdl.handle.net/" + hdlMatcher.group(2);
        }
        return baseurl;
    }

    private FairResponseModel pidValidation(String value) {

        System.out.println("your PID is " + value);

        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value == null || value.length() < 1) {
            fairResponseModel.setMessage("PID is missing.");
            fairResponseModel.setIsValid(false);

            return fairResponseModel;
        }


        String baseUrl = getPidValidatorUrl(value);

        if (baseUrl == null) {
            fairResponseModel.setMessage("PID exists but is not valid.");
            fairResponseModel.setIsValid(false);

            return fairResponseModel;
        }

        URL url = null;
        HttpURLConnection huc = null;
        int responseCode = 0;

        try {
            url = new URL(baseUrl);
            huc = (HttpURLConnection) url.openConnection();
            huc.setRequestMethod("GET");
            huc.setConnectTimeout(500);
            responseCode = huc.getResponseCode();


            monitor.info("Response code from url " + baseUrl + " is " + responseCode);

            while (responseCode == HttpURLConnection.HTTP_MOVED_PERM ||
                    responseCode == HttpURLConnection.HTTP_MOVED_TEMP ||
                    responseCode == HttpURLConnection.HTTP_SEE_OTHER) {
                String location = huc.getHeaderField("Location");
                huc = (HttpURLConnection) new URL(location).openConnection();
                huc.setRequestMethod("GET");
                huc.setConnectTimeout(500);
                responseCode = huc.getResponseCode();
                monitor.info(baseUrl + " redirected to " + location);
            }


        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        // check response status code
        if (HttpURLConnection.HTTP_OK == responseCode) {
            fairResponseModel.setMessage("PID is valid");
            fairResponseModel.setIsValid(true);
        } else {
            fairResponseModel.setMessage("PID exists but is not valid.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel licenseValidation(String value) {
        // https://github.com/spdx/license-list-data
        // https://github.com/tbouron/spdx-license-checker
        FairResponseModel fairResponseModel = new FairResponseModel();
        if (value != null && value.length() > 0) {
            boolean isValid = SpdxLicense.isValidName(value);
            fairResponseModel.setIsValid(isValid);
            if (isValid) {
                fairResponseModel.setMessage("License is valid.");
            } else {
                fairResponseModel.setMessage("License exists but is not valid.");
            }
        } else {
            fairResponseModel.setMessage("License is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;

    }

    private FairResponseModel publisherValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();
        if (value != null && value.length() > 0) {
            if (value != null && value.length() > 1) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage("Publisher is valid");
            } else {
                fairResponseModel.setIsValid(false);
                fairResponseModel.setMessage("Publisher exists but is not valid");
            }
        } else {
            fairResponseModel.setMessage("Publisher is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel titleValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();
        if (value != null && value.length() > 0) {
            if (value != null && value.length() > 1) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage("Title is valid.");
            } else {
                fairResponseModel.setIsValid(false);
                fairResponseModel.setMessage("Title exists but is not valid");
            }
        } else {
            fairResponseModel.setMessage("Title is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel dateValidation(String value, String type) {

        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() > 0) {

            if (validateJavaDate(value, "MM/dd/yyyy")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "MM-dd-yyyy")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "MM.dd.yyyy")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "dd.MM.yyyy")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "dd-MM-yy")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "dd-MM-yyyy")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "yyyy-MM-dd")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "yyyy-MM-dd HH:mm:ss")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "yyyy-MM-dd HH:mm:ss.SSS")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "yyyy-MM-dd HH:mm:ss.SSSZ")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (validateJavaDate(value, "EEEEE MMMMM yyyy HH:mm:ss.SSSZ")) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage(type + " date is valid.");
            } else if (value.indexOf("T") > -1) {
                String dt = value.substring(0, value.indexOf("T"));

                if (validateJavaDate(dt, "yyyy-MM-dd")) {
                    fairResponseModel.setIsValid(true);
                    fairResponseModel.setMessage(type + " date is valid.");
                } else {
                    fairResponseModel.setIsValid(false);
                    fairResponseModel.setMessage(type + " date exists but not valid.");
                }

            } else {
                fairResponseModel.setMessage(type + " date exists but not valid.");
                fairResponseModel.setIsValid(false);
            }
        } else {
            fairResponseModel.setMessage(type + " date is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel descriptionValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() > 0) {
            if (value != null && value.length() > 1) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage("description is valid.");
            } else {
                fairResponseModel.setIsValid(false);
                fairResponseModel.setMessage("description exists but not valid");
            }
        } else {
            fairResponseModel.setMessage("description is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    // data identifier is missing as core metadata
    private FairResponseModel keywordValidation(String[] values) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (values != null && values.length > 0) {

            boolean temp = true;
            for (int i = 0; i < values.length; i++) {
                if (values[i] == null || values.length < 1) {
                    temp = false;
                }
            }

            if (temp) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage("keywords are valid.");
            } else {
                fairResponseModel.setIsValid(false);
                fairResponseModel.setMessage("keywords exist but not valid");
            }
        } else {
            fairResponseModel.setMessage("keywords are missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel endpointValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() < 1) {
            fairResponseModel.setMessage("data endpoint is missing.");
            fairResponseModel.setIsValid(false);

            return fairResponseModel;
        }

        URL url = null;
        HttpURLConnection huc = null;
        int responseCode = 0;

        try {
            url = new URL(value);
            huc = (HttpURLConnection) url.openConnection();
            huc.setRequestMethod("HEAD");
            responseCode = huc.getResponseCode();

        } catch (Exception e) {
            // TODO Auto-generated catch block
            fairResponseModel.setMessage("Something went wrong while connecting with data endpoint." + e.toString());
            fairResponseModel.setIsValid(false);
        }

        if (HttpURLConnection.HTTP_OK == responseCode) {
            fairResponseModel.setMessage("data endpoint exists and is valid.");
            fairResponseModel.setIsValid(true);
        } else {
            fairResponseModel.setMessage("data endpoint is not found.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel dataAccessLevelValidation(String value) {
        // https://op.europa.eu/en/web/eu-vocabularies/at-dataset/-/resource/dataset/access-right
        String[] dataAccessLevelList = {"PUBLIC", "RESTRICTED", "PRIVATE", "NON_PUBLIC", "SENSITIVE", "CONFIDENTIAL"};

        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() > 0) {
            if (Arrays.asList(dataAccessLevelList).contains(value.toUpperCase())) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage("Data access level is valid");
            } else {
                fairResponseModel.setIsValid(false);
                fairResponseModel.setMessage("Data access level exists but not valid");
            }
        } else {
            fairResponseModel.setMessage("Data access level is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;

    }

    private FairResponseModel mimeTypeValidation(String mimeType) {

        FairResponseModel fairResponseModel = new FairResponseModel();

        if (mimeType == null || mimeType.length() < 1) {
            fairResponseModel.setMessage("mimeType is missing");
            fairResponseModel.setIsValid(false);
            return fairResponseModel;
        }

        System.out.println("mimeType :" + mimeType);
        fairResponseModel.setMessage("file format is valid - " + mimeType);
        fairResponseModel.setIsValid(true);

        return fairResponseModel;

    }

    private FairResponseModel languageValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() > 0) {
            fairResponseModel.setIsValid(true);
            fairResponseModel.setMessage("language is valid.");

        } else {
            fairResponseModel.setMessage("language is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel contactValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() > 0) {
            fairResponseModel.setIsValid(true);
            fairResponseModel.setMessage("contact is valid.");

        } else {
            fairResponseModel.setMessage("contact is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel byteSizeValidation(Integer value) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value != 0) {
            if (value == (int) value) {
                fairResponseModel.setIsValid(true);
                fairResponseModel.setMessage("byteSize is valid.");
            } else {
                fairResponseModel.setMessage(" byteSize is not in correct format.");
                fairResponseModel.setIsValid(false);
            }
        } else {
            fairResponseModel.setMessage("byteSize is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    private FairResponseModel versionValidation(String value) {
        FairResponseModel fairResponseModel = new FairResponseModel();

        if (value != null && value.length() > 0) {
            fairResponseModel.setIsValid(true);
            fairResponseModel.setMessage("version is valid.");

        } else {
            fairResponseModel.setMessage("version is missing.");
            fairResponseModel.setIsValid(false);
        }

        return fairResponseModel;
    }

    public static boolean validateJavaDate(String strDate, String format) {
        /*
         * Set preferred date format, For example MM-dd-yyyy, MM.dd.yyyy,dd.MM.yyyy etc.
         */
        // SimpleDateFormat sdfrmt = new SimpleDateFormat("MM/dd/yyyy");
        SimpleDateFormat sdfrmt = new SimpleDateFormat(format);
        sdfrmt.setLenient(false);

        try {
            Date javaDate = sdfrmt.parse(strDate);
            System.out.println(strDate + " is valid date format");
        } /* Date format is invalid */ catch (ParseException e) {
            System.out.println(strDate + " is Invalid Date format");
            return false;
        }
        return true;
    }

}

